# Copyright (C) 2021 LinDev
#
# Author: Bernhard Rosenkraenzer <bero@lindev.ch>
#
# This file is part of cloud-init. See LICENSE file for license information.

from cloudinit.distros import fedora


class Distro(fedora.Distro):
    pass
